#include "tilda.h"
#include "tilda-controller.h"
#include "tilda-controller-dbus-glue.h"
#include "tilda-window.h"

/**
 * Register this object with DBus, so it can be interacted with!
 */
static void
tilda_controller_dbus_register_object (TildaController *self)
{
	debug_enter  ();
	debug_assert (TILDA_IS_CONTROLLER(self));

	static const gchar object_path[] = "/net/sourceforge/Tilda";

	dbus_g_connection_register_g_object (dbus_connection, object_path, G_OBJECT(self));
}

/**
 * Find the next free window number, so it can be used by a TildaWindow
 */
static gint
tilda_controller_find_next_free_window_number (TildaController *self)
{
	debug_enter  ();
	debug_assert (TILDA_IS_CONTROLLER(self));

	gint i, j;
	gboolean found;

	for (i=0; i<INT_MAX; ++i)
	{
		found = FALSE;

		for (j=0; j<self->windows->len; ++j)
		{
			TildaWindow *tw = g_ptr_array_index (self->windows, j);

			if (tw->number == i)
			{
				found = TRUE;
				break;
			}
		}

		if (!found)
			return i;
	}

	return 0;
}

gboolean
tilda_controller_add_window (TildaController *self)
{
	debug_enter  ();
	debug_assert (TILDA_IS_CONTROLLER(self));

	TildaWindow *ret;
	gint number;

	number = tilda_controller_find_next_free_window_number (self);
	ret = g_object_new (TILDA_TYPE_WINDOW,
						"number", number,
						"controller", G_OBJECT(self),
						NULL);

	g_ptr_array_add (self->windows, ret);

	debug_printf ("Adding window: 0x%x (number %d of %d)\n", ret, ret->number, self->windows->len);
	return TRUE;
}

gboolean
tilda_controller_delete_window (TildaController *self, gint window_number)
{
	debug_enter  ();
	debug_assert (TILDA_IS_CONTROLLER(self));
	debug_assert (window_number >= 0);

	gint i;
	TildaWindow *win;

	for (i=0; i<self->windows->len; ++i)
	{
		win = g_ptr_array_index (self->windows, i);

		if (win->number == window_number)
		{
			debug_printf ("Deleting TildaWindow 0x%x (number %d of %d)\n",
						  win, win->number, self->windows->len);
			g_ptr_array_remove_index (self->windows, i);
			g_object_unref (G_OBJECT(win));

			if (self->windows->len == 0)
			{
				debug_printf ("No windows left, exiting...\n");

				/* We get unref'd in main() */
				gtk_main_quit ();
			}

			/* We were able to delete the window */
			return TRUE;
		}
	}

	/* There must have been no window to delete */
	return FALSE;
}

gboolean
tilda_controller_quit (TildaController *self, GError **error)
{
	debug_enter  ();
	debug_assert (TILDA_IS_CONTROLLER(self));

	/* Not much left but to quit, since we get unref'd in main() */
	gtk_main_quit ();

	return TRUE;
}

/*******************************************************************************
 * GObject code below... it is doubtful you'll need to make big changes :)
 ******************************************************************************/

static GObjectClass *parent_class = NULL;

static void
tilda_controller_instance_init (GTypeInstance *instance,
								gpointer       g_class)
{
	debug_enter ();

	TildaController *self = (TildaController *) instance;

	self->dispose_has_run = FALSE;
	self->windows = g_ptr_array_new ();
}

static void
tilda_controller_set_property (GObject      *object,
							   guint         property_id,
							   const GValue *value,
							   GParamSpec   *pspec)
{
	TildaController *self = (TildaController *) self;

	switch (property_id)
	{
		default:
			/* We don't have any other properties */
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
			break;
	}
}

static void
tilda_controller_get_property (GObject    *object,
							   guint       property_id,
							   GValue     *value,
							   GParamSpec *pspec)
{
	TildaController *self = (TildaController *) object;

	switch (property_id)
	{
		default:
			/* We don't have any other properties */
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
			break;
	}
}

static GObject *
tilda_controller_constructor (GType                  type,
							  guint                  n_construct_properties,
							  GObjectConstructParam *construct_properties)
{
	debug_enter ();

	GObject *obj;
	TildaControllerClass *klass;
	TildaController *self;

	/* Invoke the parent constructor */
	klass = TILDA_CONTROLLER_CLASS (g_type_class_peek (TILDA_TYPE_CONTROLLER));
	obj = parent_class->constructor (type,
									 n_construct_properties,
									 construct_properties);

	/* The object is ready, and all constructor-time properties have been set.
	 * Have fun! */
	self = TILDA_CONTROLLER(obj);

	/* Register this object with DBus */
	tilda_controller_dbus_register_object (self);

	/* Add a window -- FIXME: the number should be configurable */
	tilda_controller_add_window (self);

	return obj;
}

static void
tilda_controller_dispose (GObject *obj)
{
	debug_enter ();

	TildaController *self = (TildaController *) obj;

	/* We must only run dispose once ... */
	if (self->dispose_has_run)
		return;

	self->dispose_has_run = TRUE;

	/*
	 * In dispose, you are supposed to free all types referenced from this
	 * object which might themselves hold a reference to self. Generally,
	 * the most simple solution is to unref all members on which you own a
	 * reference.
	 */
	g_ptr_array_foreach (self->windows, g_object_unref, NULL);

	/* Chain up to the parent class */
	G_OBJECT_CLASS (parent_class)->dispose (obj);
}

static void
tilda_controller_finalize (GObject *obj)
{
	debug_enter ();

	TildaController *self = (TildaController *) obj;

	/*
	 * Here, complete the object's destruction.
	 * You might not need to do much more than
	 * g_free() any primitives.
	 */

	/* Chain up to the parent class */
	G_OBJECT_CLASS (parent_class)->finalize (obj);
}

static void
tilda_controller_class_init (gpointer g_class,
					   gpointer g_class_data)
{
	debug_enter ();

	GObjectClass *gobject_class = G_OBJECT_CLASS (g_class);
	TildaControllerClass *klass = TILDA_CONTROLLER_CLASS (g_class);
	GParamSpec *pspec;

	gobject_class->set_property = tilda_controller_set_property;
	gobject_class->get_property = tilda_controller_get_property;
	gobject_class->dispose = tilda_controller_dispose;
	gobject_class->finalize = tilda_controller_finalize;
	gobject_class->constructor = tilda_controller_constructor;

	parent_class = g_type_class_peek_parent (klass);

	/* Add properties here */

	/* Hook the TildaController type into DBus */
	dbus_g_object_type_install_info (tilda_controller_get_type(),
									 &dbus_glib_tilda_controller_object_info);
}

GType tilda_controller_get_type (void)
{
	static GType type = 0;

	if (type == 0)
	{
		static const GTypeInfo info = {
			sizeof (TildaControllerClass),
			NULL,	/* base_init */
			NULL,	/* base_finalize */
			tilda_controller_class_init,	/* class_init */
			NULL,	/* class_finalize */
			NULL,	/* class_data */
			sizeof (TildaController),
			0,		/* n_preallocs */
			tilda_controller_instance_init,	/* instance_init */
		};

		type = g_type_register_static (G_TYPE_OBJECT,
									   "TildaControllerType",
									   &info,
									   0);
	}

	return type;
}

/* vim: set ts=4 sts=4 sw=4 noet tw=112: */
